import bulwark.checks as ck
import numpy as np
from bayes_window.generative_models import generate_fake_spikes


def test_generate_fake_spikes():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes()
    df = df.set_index(list(index_cols))
    assert abs(firing_rates.sel(Stim=1).isel(Neuron=0, Mouse=0) -
               firing_rates.sel(Stim=0).isel(Neuron=0, Mouse=0)
               ) < 1

    assert (firing_rates.sel(Stim=1).isel(Neuron=0, Mouse=-1) -
            firing_rates.sel(Stim=0).isel(Neuron=0, Mouse=-1)
            ) > 1

    assert abs(firing_rates.sel(Stim=1).isel(Neuron=-1, Mouse=-1) -
               firing_rates.sel(Stim=0).isel(Neuron=-1, Mouse=-1)
               ) < 1

    df.pipe(ck.has_no_nans)
    df_monster.pipe(ck.has_no_nans)



import warnings
from importlib import reload

import altair as alt
import numpy as np
import pandas as pd
import statsmodels.formula.api as sm
from sklearn.preprocessing import LabelEncoder

from bayes_window import models
from bayes_window import utils
from bayes_window import visualization
from bayes_window.fitting import fit_numpyro
from bayes_window.model_comparison import compare_models
from bayes_window.visualization import plot_posterior

reload(visualization)
le = LabelEncoder()


def combined_condition(df, levels):
    # String-valued combined condition
    # df['combined_condition'] = utils.df_index_compress(df, index_columns=self.levels)[1]
    df['combined_condition'] = df[levels[0]].astype('str')
    for level in levels[1:]:
        df['combined_condition'] += df[level].astype('str')
    data = df.copy()
    # Transform conditions to integers as required by numpyro:
    data['combined_condition'] = le.fit_transform(df['combined_condition'])
    # Transform conditions to integers as required by numpyro:
    key = dict()
    for level in levels:
        data[level] = le.fit_transform(data[level])
        # Keep key for later use
        # TODO I don't think this key is currently used for plotting, is it? Test with mouse names
        key[level] = dict(zip(range(len(le.classes_)), le.classes_))
    return data, key


class BayesWindow:
    def __init__(self,
                 df: pd.DataFrame,
                 y: str,
                 treatment: str,
                 condition: str or list = None,
                 group: str = None,
                 detail=':O'
                 ):
        assert y in df.columns
        assert treatment in df.columns
        if group:
            assert group in df.columns
        self.treatment = treatment  # if type(treatment)=='list' else [treatment]  # self.levels[2]
        self.group = group  # if type(group)=='list' else [group] # self.levels[1]  # Eg subject
        self.condition = condition if type(condition) == list else [condition]
        if self.condition[0]:
            assert self.condition[0] in df.columns
        self.levels = utils.parse_levels(self.treatment, self.condition, self.group)
        self.data, self._key = combined_condition(df, self.levels)
        self.detail = detail
        self.y = y

        # Preallocate attributes:
        self.b_name = None  # Depends on model we'll fit
        self.do_make_change = None  # Depends on plotting input
        self.add_data = None  # We'll use this in plotting
        self.independent_axes = None
        self.data_and_posterior = None
        self.posterior = None
        self.trace = None
        self.model = None

    def fit_anova(self):
        from statsmodels.stats.anova import anova_lm
        if self.group:
            # Average over group:
            data = self.data.groupby([self.group, self.treatment]).mean().reset_index()
        else:
            data = self.data
        # dehumanize all columns and variable names for statsmodels:
        [data.rename({col: col.replace(" ", "_")}, axis=1, inplace=True) for col in data.columns]
        self.y = self.y.replace(" ", "_")
        formula = f'{self.y}~{self.treatment}'
        lm = sm.ols(formula, data=data).fit()
        print(f'{formula}\n {anova_lm(lm, typ=2)}')
        return anova_lm(lm, typ=2)['PR(>F)'][self.treatment] < 0.05

    def fit_lme(self, add_data=False, do_make_change='divide'):
        # model = MixedLM(endog=self.data[self.y],
        #                 exog=self.data[self.condition],
        #                 groups=self.data[self.group],
        #                 # exog_re=exog.iloc[:, 0]
        #                 )
        # dehumanize all columns and variable names for statsmodels:
        [self.data.rename({col: col.replace(" ", "_")}, axis=1, inplace=True) for col in self.data.columns]
        self.y = self.y.replace(" ", "_")
        self.group = self.group.replace(" ", "_")
        self.treatment = self.treatment.replace(" ", "_")
        self.do_make_change = do_make_change
        include_condition = False
        if self.condition[0]:
            if len(self.data[self.condition[0]].unique()) > 1:
                include_condition = True
        if include_condition:
            if len(self.condition) > 1:
                raise NotImplementedError(f'conditions {self.condition}')
                # This would need a combined condition dummy variable and an index of condition in patsy:
                # formula = f"{self.y} ~ 1+ {self.condition}(condition_index) | {self.treatment}"
            self.condition[0] = self.condition[0].replace(" ", "_")
            formula = f"{self.y} ~ 1+ {self.condition[0]} | {self.treatment}"
        else:
            formula = f"{self.y} ~ 1 + {self.treatment}"
        if self.group:
            formula += f' + (1 | {self.group})'
        print(f'Using formula {formula}')
        result = sm.mixedlm(formula,
                            self.data,
                            groups=self.data[self.group]).fit()
        res = result.summary().tables[1]
        res = res.iloc[:-1]
        try:
            res = res.astype(float)  # [['P>|z|', 'Coef.', '[0.025', '0.975]']]
        except Exception as e:
            warnings.warn(f'somehow LME failed to estimate CIs for one or more variables. Replacing with nan:'
                          f' {e} \n=>\n {res}')
            res.replace({'': np.nan}).astype(float)
        res = res.rename({'P>|z|': 'p',
                          'Coef.': 'center interval',
                          '[0.025': 'higher interval',
                          '0.975]': 'lower interval'}, axis=1)
        self.posterior = res
        if add_data and self.condition[0]:
                raise NotImplementedError("I don't understand if there is a way to get separate estimates "
                                          "of slope per condition in LME, or do you just get an effect size estimate??")
                # like in hdi2df:
                from utils import fill_row
                rows = [fill_row(group_val, rows, res, group_name=self.condition[0])
                        for group_val, rows in self.data.groupby([self.condition[0]])
                        ]
                self.data_and_posterior = pd.concat(rows)
        elif add_data:
            # like in hdi2df_one_condition():
            self.data_and_posterior = self.data.copy()
            for col in ['lower interval', 'higher interval', 'center interval']:
                self.data_and_posterior.insert(self.data.shape[1],
                                               col,
                                               res.loc[self.treatment, col])

        return self

    def fit_conditions(self, model=models.model_single, add_data=True):

        self.model = model
        self.b_name = 'mu_per_condition'
        # Estimate model
        self.trace = fit_numpyro(y=self.data[self.y].values,
                                 condition=self.data['combined_condition'].values,
                                 model=model,
                                 ).posterior

        # Add data back
        if add_data:
            self.data_and_posterior, self.trace = utils.add_data_to_posterior(df_data=self.data,
                                                                              posterior=self.trace,
                                                                              y=self.y,
                                                                              fold_change_index_cols=self.levels[:3],
                                                                              treatment_name=self.levels[0],
                                                                              b_name=self.b_name,
                                                                              posterior_index_name='combined_condition',
                                                                              do_mean_over_trials=False,
                                                                              do_make_change=False,
                                                                              add_data=add_data
                                                                              )
        return self

    def fit_slopes(self, model=models.model_hierarchical, do_make_change='subtract',
                   fold_change_index_cols=None, do_mean_over_trials=True, add_data: bool = True, **kwargs):
        # TODO case with no group_name
        if do_make_change not in ['subtract', 'divide']:
            raise ValueError(f'do_make_change should be subtract or divide, not {do_make_change}')

        self.b_name = 'b_stim_per_condition'
        self.do_make_change = do_make_change
        self.add_data = add_data  # We'll use this in plotting
        self.model = model
        # TODO handle no-group case
        if fold_change_index_cols is None:
            # TODO case with no plot_index_cols should include any multiindex?
            fold_change_index_cols = self.levels  # [-1]
        if not self.condition[0]:
            warnings.warn('Condition was not provided. Assuming there is no additional condition, just treatment')
            self.condition[0] = 'dummy_condition'
            self.data.insert(self.data.shape[-1] - 1, 'dummy_condition', np.ones(self.data.shape[0]))
            fold_change_index_cols.append('dummy_condition')
        if not self.condition[0] in fold_change_index_cols:
            fold_change_index_cols.extend(self.condition)
        try:
            self.trace = fit_numpyro(y=self.data[self.y].values,
                                     treatment=self.data[self.treatment].values,
                                     condition=self.data[self.condition[0]].values,
                                     group=self.data[self.group].values,
                                     progress_bar=False,
                                     model=model,
                                     n_draws=1000, num_chains=1,
                                     **kwargs)
        except TypeError as e:
            # assert that model() has kwarg stim, because this is slopes
            raise KeyError(f'Does your model {model} have "stim" argument? You asked for slopes!{e}')
        if add_data:
            # Add data back
            # TODO posterior_index_name=self.condition[0] will not work if need combined_condition
            df_result, self.trace.posterior = utils.add_data_to_posterior(df_data=self.data,
                                                                          posterior=self.trace.posterior,
                                                                          y=self.y,
                                                                          fold_change_index_cols=fold_change_index_cols,
                                                                          treatment_name=self.treatment,
                                                                          b_name=self.b_name,
                                                                          posterior_index_name=self.condition[0],
                                                                          do_make_change=do_make_change,
                                                                          do_mean_over_trials=do_mean_over_trials,
                                                                          add_data=self.add_data
                                                                          )
        else:  # Just convert posterior to dataframe
            from bayes_window.utils import trace2df
            # TODO we add data regardless. Is there a way to not use self.data?
            df_result, self.trace.posterior = trace2df(self.trace.posterior,
                                                       self.data, b_name=self.b_name,
                                                       posterior_index_name=self.condition[0])

        # Back to human-readable labels
        [df_result[col].replace(self._key[col], inplace=True) for col in self._key.keys()
         if (not col == self.treatment) and (col in df_result)]
        self.data_and_posterior = df_result
        return self

    def plot_posteriors_slopes(self, x=':O', color=':O', add_box=True, independent_axes=False, **kwargs):
        # Set some options
        self.independent_axes = independent_axes
        x = x or self.levels[-1]
        if x[-2] != ':':
            x += ':O'
        color = color or self.levels[0]

        # Plot posterior
        if self.data_and_posterior is not None:
            add_data = self.add_data
            base_chart = alt.Chart(self.data_and_posterior)
            chart_p = plot_posterior(title=f'{self.y}',
                                     x=x,
                                     base_chart=base_chart,
                                     do_make_change=self.do_make_change)
        else:
            base_chart = alt.Chart(self.data)
            add_data = True  # Otherwise nothing to do

        if add_data:
            assert self.data_and_posterior is not None
            chart_d = visualization.plot_data(x=x, y=f'{self.y} diff', color=color, add_box=add_box,
                                              base_chart=base_chart)
            self.chart = chart_d + chart_p
        else:
            self.chart = chart_p
        if independent_axes:
            self.chart = self.chart.resolve_scale(y='independent')
        return self.chart

    # TODO plot_posteriors_slopes and plot_posteriors_no_slope can be one
    def plot_posteriors_no_slope(self,
                                 x=None,
                                 add_data=False,
                                 independent_axes=True,
                                 color=None,
                                 detail=':O',
                                 **kwargs):
        self.independent_axes = independent_axes
        x = x or self.treatment
        detail = detail or self.detail
        color = color or self.condition[0]
        # TODO default for detail
        if self.data_and_posterior is not None:
            base_chart = alt.Chart(self.data_and_posterior)
            # Plot posterior
            chart_p = visualization.plot_posterior(x=x,
                                                   do_make_change=False,
                                                   add_data=add_data,
                                                   # title=f'{self.y} estimate', #TODO uncomment
                                                   base_chart=base_chart,
                                                   **kwargs
                                                   )
            if not add_data:  # done
                self.chart = chart_p
        else:
            add_data = True  # Otherwise nothing to do
            base_chart = alt.Chart(self.data)

        if add_data:
            # Make data plot:
            chart_d = visualization.plot_data_slope_trials(x=x,
                                                           y=self.y,
                                                           color=color,
                                                           detail=detail,
                                                           base_chart=base_chart)

            if self.data_and_posterior is None:
                self.chart = chart_d  # we're done
            else:
                self.chart = chart_p + chart_d

        return self.chart

    def plot(self, **kwargs):
        # Convenience function
        if not self.b_name:
            warnings.warn('No model has been fit. Defaulting to plotting "slopes" for data. Use .plot_slopes'
                          'or .plot_posteriors_no_slope to be explicit ')
            return visualization.plot_data(self.data, x=self.levels[0], y=self.y,
                                           color=self.levels[1] if len(self.levels) > 1 else None,
                                           **kwargs)

        if self.b_name == 'b_stim_per_condition':
            return BayesWindow.plot_posteriors_slopes(self, **kwargs)
        elif self.b_name == 'mu_per_condition':
            return BayesWindow.plot_posteriors_no_slope(self, **kwargs)

    def facet(self, width=50, height=60, **kwargs):
        assert ('row' in kwargs) or ('column' in kwargs), 'Give facet either row, or column'
        if self.independent_axes is None:
            # TODO let's not force users to plot. have a sensible default
            raise RuntimeError('Plot first, then you can use facet')
        if self.independent_axes:
            self.facetchart = visualization.facet(self.chart, width=width, height=height, **kwargs)
        else:
            self.facetchart = self.chart.properties(width=width, height=height).facet(**kwargs)
        return self.facetchart

    def plot_model_quality(self, var_names=None):
        import arviz as az
        assert hasattr(self, 'trace'), 'Run bayesian fitting first!'
        az.plot_trace(self.trace, var_names=var_names, show=True)

    def explore_models(self, parallel=True):
        if self.b_name is None:
            raise ValueError('Fit a model first')
        elif self.b_name == 'mu_per_condition':
            return compare_models(df=self.data,
                                  models={
                                      'no_condition': self.model,
                                      'full_normal': self.model,
                                      'full_student': self.model,
                                      'full_lognogmal': self.model,
                                  },
                                  extra_model_args=[
                                      {'condition': None},
                                      {'condition': self.condition},
                                      {'condition': self.condition},
                                      {'condition': self.condition},
                                  ],
                                  y=self.y,
                                  parallel=True
                                  )

        elif self.b_name == 'b_stim_per_condition':
            return compare_models(df=self.data,
                                  models={
                                      'full_normal': self.model,
                                      'no_condition': self.model,
                                      'no_condition_or_treatment': self.model,
                                      'no-treatment': self.model,
                                      'no_group': self.model,
                                      'full_student': self.model,
                                      'full_lognogmal': self.model,
                                  },
                                  extra_model_args=[
                                      {'treatment': self.treatment, 'condition': self.condition, 'group': self.group},
                                      {'treatment': self.treatment, 'condition': None},
                                      {'treatment': None, 'condition': None},
                                      {'treatment': None, 'condition': self.condition},
                                      {'treatment': self.treatment, 'condition': self.condition, 'group': None},
                                      {'treatment': self.treatment, 'condition': self.condition, 'group': self.group,
                                       'dist_y': 'student', },
                                      {'treatment': self.treatment, 'condition': self.condition, 'group': self.group,
                                       'dist_y': 'lognormal'},
                                  ],
                                  y=self.y,
                                  parallel=parallel
                                  )



from pathlib import Path

from bayes_window import models
from bayes_window.generative_models import *
from bayes_window.visualization import plot_posterior
from bayes_window.workflow import BayesWindow

trans = LabelEncoder().fit_transform


def test_slopes_dont_make_change():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=5,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=7,
                                                                    mouse_response_slope=16)
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    try:
        bw.fit_slopes(add_data=False, model=models.model_hierarchical, do_make_change=False,
                      fold_change_index_cols=('stim', 'mouse', 'neuron'))
    except ValueError:
        pass


def test_fit_lme():
    df, df_monster, index_cols, _ = generate_fake_lfp(n_trials=25)
    bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
    bw.fit_lme(add_data=False, )


def test_fit_lme_w_data():
    df, df_monster, index_cols, _ = generate_fake_lfp(n_trials=25)

    bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
    bw.fit_lme(add_data=True, do_make_change='divide')


def test_fit_lme_w_data_condition():
    df, df_monster, index_cols, _ = generate_fake_spikes(n_trials=25)

    bw = BayesWindow(df, y='isi', treatment='stim', group='mouse', condition='neuron')
    try:
        bw.fit_lme(add_data=True, do_make_change='divide')
    except NotImplementedError:
        pass


def test_estimate_posteriors():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron_code', group='mouse', )
    bw.fit_conditions(model=models.model_single, add_data=False)

    chart = bw.plot(x='stim:O', column='neuron', row='mouse', )
    chart.display()
    chart = bw.plot(x='stim:O', column='neuron_code', row='mouse_code', )
    chart.display()


def test_estimate_posteriors_data_overlay():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron_code', group='mouse')
    bw.fit_conditions(model=models.model_single, add_data=False)
    chart = bw.plot(x='stim:O', independent_axes=False,
                    column='neuron_code', row='mouse_code')
    chart.display()


def test_estimate_posteriors_data_overlay_indep_axes():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron_code', group='mouse')
    bw.fit_conditions(model=models.model_single, add_data=True, )

    chart = bw.plot(x='stim:O', independent_axes=True,
                    column='neuron_code', row='mouse_code')
    chart.display()


def test_plot():
    from bayes_window.workflow import BayesWindow

    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    chart = BayesWindow(df, y='isi', treatment='stim').plot()
    chart.display()


def test_estimate_posteriors_slope():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron_code', group='mouse', )
    bw.fit_slopes(models.model_hierarchical, add_data=False)

    chart = bw.plot(x='neuron_code', column='neuron_code', row='mouse')
    chart.display()
    chart = bw.plot(x='neuron_code', column='neuron_code', row='mouse_code')
    chart.display()


def test_estimate_posteriors_slope_strengths():
    df = generate_spikes_stim_types(mouse_response_slope=3,
                                    n_trials=2,
                                    n_neurons=3,
                                    n_mice=4,
                                    dur=2, )
    if 1:  # TODO this will be hairy
        bw = BayesWindow(df, y='isi', treatment='stim', condition=['neuron_code', 'stim_strength'], group='mouse', )
        bw.fit_slopes(model=models.model_hierarchical, do_mean_over_trials=False, fold_change_index_cols=None,
                      add_data=False)

        chart = bw.plot(x='neuron_code', column='neuron_code', row='mouse')
        chart.display()
        chart = bw.plot(x='neuron_code', column='neuron_code', row='mouse_code')
        chart.display()


def test_estimate_posteriors_data_overlay_slope():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron_code', group='mouse')
    bw.fit_slopes(model=models.model_hierarchical, add_data=False)
    chart = bw.plot_posteriors_slopes(independent_axes=False)
    chart.display()
    bw.facet(column='neuron_code', row='mouse_code')
    chart.display()


def test_estimate_posteriors_data_overlay_indep_axes_slope():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron_code', group='mouse')
    bw.fit_slopes(model=models.model_hierarchical, add_data=True)
    chart = bw.plot_posteriors_no_slope(independent_axes=True)
    chart.display()
    chart = bw.facet(column='neuron_code', row='mouse')
    chart.display()


def test_plot_no_slope_data_only():
    from bayes_window.workflow import BayesWindow

    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    chart = BayesWindow(df, y='isi', treatment='stim').plot_posteriors_no_slope()
    chart.display()


def test_plot_slope_data_only():
    from bayes_window.workflow import BayesWindow

    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    chart = BayesWindow(df, y='isi', treatment='stim').plot_posteriors_no_slope()
    chart.display()


def test_fit_conditions():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    # TODO combined condition here somehow
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_conditions(add_data=True)


def test_fit_slopes():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hierarchical, )


def test_plot_slopes():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hierarchical, )
    bw.plot()


def test_plot_posteriors_no_slope():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hierarchical, )
    bw.plot_posteriors_slopes()


def test_plot_generic():
    # Slopes:
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hierarchical, )
    bw.plot()
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=False, model=models.model_hierarchical, )
    bw.plot()
    # conditions:
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_conditions(model=models.model_single)
    bw.plot()


def test_facet():
    # Slopes:
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hierarchical, )
    bw.plot().facet('neuron', width=40)

    # conditions:
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_conditions(model=models.model_single)
    bw.plot().facet('neuron', width=40)


def test_single_condition_withdata():
    df, df_monster, index_cols, _ = generate_fake_lfp()
    bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hier_stim_one_codition,
                  do_make_change='divide', dist_y='normal')
    plot_posterior(df=bw.data_and_posterior, title=f'Log power', ).display()
    bw.plot_posteriors_slopes(add_box=True, independent_axes=True).display()

    # Without data again
    bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
    bw.fit_slopes(add_data=False, model=models.model_hier_stim_one_codition,
                  do_make_change='divide', dist_y='normal')
    plot_posterior(df=bw.data_and_posterior, title=f'Log power', ).display()
    bw.plot_posteriors_slopes(add_box=True, independent_axes=True).display()

    # With data again
    bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hier_stim_one_codition,
                  do_make_change='divide', dist_y='normal')
    plot_posterior(df=bw.data_and_posterior, title=f'Log power', ).display()
    bw.plot_posteriors_slopes(add_box=True, independent_axes=True).display()


def test_single_condition_nodata():
    df, df_monster, index_cols, _ = generate_fake_lfp()
    bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
    bw.fit_slopes(add_data=False, model=models.model_hier_stim_one_codition,
                  do_make_change='divide', dist_y='normal')
    plot_posterior(df=bw.data_and_posterior, title=f'Log power', ).display()
    bw.plot_posteriors_slopes(add_box=True, independent_axes=True).display()


def test_single_condition_nodata_dists():
    df, df_monster, index_cols, _ = generate_fake_lfp()
    for dist in ['normal', 'lognormal', 'student']:
        bw = BayesWindow(df, y='Log power', treatment='stim', group='mouse')
        bw.fit_slopes(add_data=False, model=models.model_hier_stim_one_codition,
                      do_make_change='divide', dist_y=dist)
        plot_posterior(df=bw.data_and_posterior, title=f'Log power', ).display()
        bw.plot_posteriors_slopes(add_box=True, independent_axes=True).display()


def test_explore_models():
    # Slopes:
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    bw = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')
    bw.fit_slopes(add_data=True, model=models.model_hierarchical, )
    bw.explore_models(parallel=False)
    bw.explore_models(parallel=True)


def test_chirp_data():
    df = pd.read_csv(Path('tests') / 'test_data' / 'chirp_power.csv')
    window = BayesWindow(df, y='Log power',
                         treatment='stim_on',
                         condition='Condition code',
                         group='Subject')
    window.fit_slopes(model=models.model_hierarchical, do_mean_over_trials=True,
                      fold_change_index_cols=['Condition code',
                                              'Brain region', 'Stim phase', 'stim_on', 'Fid', 'Subject', 'Inversion'], )
    window.plot_posteriors_slopes(x='Stim phase', color='Fid', independent_axes=True)


def test_chirp_data1():
    df = pd.read_csv(Path('tests') /('test_data') / 'chirp_power.csv')
    window = BayesWindow(df, y='Log power',
                         treatment='stim_on',
                         condition='Condition code',
                         group='Subject')
    window.fit_slopes(model=models.model_hierarchical, do_mean_over_trials=True,
                      fold_change_index_cols=[  # 'Condition code',
                          'Brain region', 'Stim phase', 'stim_on', 'Fid', 'Subject', 'Inversion'], )
    window.plot_posteriors_slopes(x='Stim phase', color='Fid', independent_axes=True)

def test_conditions2():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=5,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=7,
                                                                    mouse_response_slope=16)
    df.neuron = df.neuron.astype(int)
    window = BayesWindow(df, y='isi', treatment='stim', condition='neuron', group='mouse')

    window.fit_conditions(model=models.model_single, )
    assert window.y in window.data_and_posterior
    window.plot_posteriors_no_slope(x='stim:O', independent_axes=False, add_data=True);



import warnings

import arviz as az
import numpy as np
import pandas as pd
import scipy
import xarray as xr
from sklearn.preprocessing import LabelEncoder

trans = LabelEncoder().fit_transform


def level_to_data_column(level_name, kwargs):
    from collections import Iterable
    # import itertools
    # flatten = itertools.chain.from_iterable
    x = kwargs[level_name]
    if isinstance(x, Iterable) and not isinstance(x, (str, bytes)):
        if len(x) > 1:
            raise ValueError(f'Multiple conditions are not supported:{x}')
        # list(flatten(level_name))
        return kwargs[level_name][0]
    else:
        return kwargs[level_name]


def parse_levels(treatment, condition, group):
    levels = []
    if treatment:
        levels += [treatment]
    if condition[0]:
        levels += condition
    if group:
        levels += [group]
    return levels


def add_data_to_posterior(df_data,
                          posterior,
                          y=None,  # Only for fold change
                          fold_change_index_cols=None,
                          treatment_name='Event',
                          treatments=None,  # eg ('stim_on', 'stim_stop')
                          b_name='b_stim_per_condition',  # for posterior
                          posterior_index_name='',  # for posterior
                          do_make_change='subtract',
                          do_mean_over_trials=True,
                          add_data=False
                          ):
    # group_name should be conditions
    if type(fold_change_index_cols) == str:
        fold_change_index_cols = [fold_change_index_cols]
    fold_change_index_cols = list(fold_change_index_cols)
    treatments = treatments or df_data[treatment_name].drop_duplicates().sort_values().values

    assert len(treatments) == 2, f'{treatment_name}={treatments}. Should be only two instead!'
    assert do_make_change in [False, 'subtract', 'divide']
    if not (treatment_name in fold_change_index_cols):
        fold_change_index_cols.append(treatment_name)
    # if not (group_name in index_cols):
    #     index_cols.append(group_name)
    if do_mean_over_trials:
        df_data = df_data.groupby(fold_change_index_cols).mean().reset_index()
    if do_make_change:
        # Make (fold) change
        df_data, _ = make_fold_change(df_data,
                                      y=y,
                                      index_cols=fold_change_index_cols,
                                      treatment_name=treatment_name,
                                      treatments=treatments,
                                      fold_change_method=do_make_change,
                                      do_take_mean=False)
    # Convert to dataframe and fill in data:
    df_bayes, posterior = trace2df(posterior, df_data, b_name=b_name, posterior_index_name=posterior_index_name,
                                   add_data=add_data)
    return df_bayes, posterior


def fill_row(condition_val, rows, df_bayes, condition_name):
    this_hdi = df_bayes.loc[df_bayes[condition_name] == condition_val]
    assert this_hdi.shape[0] > 0, \
        f'No such value {condition_val} in {condition_name}: it"s {df_bayes[condition_name].unique()}'
    for col in ['lower interval', 'higher interval', 'center interval']:
        rows.insert(rows.shape[1] - 1, col, this_hdi[col].values.squeeze())
    return rows


def hdi2df_many_conditions(df_bayes, posterior_index_name, df_data):
    # Check
    if len(df_data[posterior_index_name].unique()) != len(df_bayes[posterior_index_name].unique()):
        raise ValueError('Groups were constructed differently for estimation and data. Cant add data for plots')
    rows = [fill_row(group_val, rows, df_bayes, posterior_index_name)
            for group_val, rows in df_data.groupby([posterior_index_name])]
    return pd.concat(rows)


def hdi2df_one_condition(df_bayes, group_name, df_data):
    df_bayes[group_name] = df_data[group_name].iloc[0]
    for col in ['lower interval', 'higher interval', 'center interval']:
        df_data.insert(df_data.shape[1], col, df_bayes[col].values.squeeze())
    return df_data


def _mode(*args, **kwargs):
    vals = scipy.stats.mode(*args, **kwargs)
    # only return the mode (discard the count)
    return vals[0].squeeze()


def xar_mode(obj, dims_to_reduce: list = None, dim=None):
    # xar_mode(trace[b_name], dims_to_reduce=['chain', 'draw']) # OR:
    # xar_mode(trace[b_name], dim='neuron')

    # note: apply always moves core dimensions to the end
    # usually axis is simply -1 but scipy's mode function doesn't seem to like that
    # this means that this version will only work for DataArrays (not Datasets)
    assert isinstance(obj, xr.DataArray)
    assert (dims_to_reduce is not None) or (dim is not None)
    axis = obj.ndim - 1
    if dims_to_reduce is None:
        dims_to_reduce = list(set(obj.dims) - {dim})
    return xr.apply_ufunc(_mode, obj,
                          input_core_dims=[dims_to_reduce],
                          # exclude_dims=set(['chain','draw']),
                          kwargs={'axis': axis}
                          )


def trace2df(trace, df_data, b_name='b_stim_per_condition', posterior_index_name='neuron', add_data=False):
    """
    # Convert to dataframe and fill in original conditions
    group name is whatever was used to index posterior
    """
    # TODO this is lazy. There may be more than one condition, need to include them all instead of combined_condition
    if f'{b_name}_dim_0' in trace:
        trace = trace.rename({f'{b_name}_dim_0': posterior_index_name})
    if f'a_subject_dim_0' in trace:
        trace = trace.rename({f'a_subject_dim_0': 'subject'})
    if add_data and (df_data[posterior_index_name].dtype != 'int'):
        warnings.warn(
            f"Was {posterior_index_name} a string? It's safer to recast it as integer. I'll try to do that...")
        df_data[posterior_index_name] = df_data[posterior_index_name].astype(int)

    hdi = az.hdi(trace)[b_name]

    # from scipy.stats import mode
    max_a_p = xar_mode(trace[b_name], dims_to_reduce=['chain', 'draw'])
    # Or mean trace[b_name].mean(['chain', 'draw']).values,
    if hdi.ndim == 1:
        mean = xr.DataArray([max_a_p],
                            coords={'hdi': ["center"], },
                            dims='hdi')
        df_bayes = xr.concat([hdi, mean], 'hdi').to_dataframe().reset_index()
        df_bayes = df_bayes.pivot_table(columns='hdi').reset_index(drop=True)
        if not df_bayes.columns.str.contains('interval').any():
            # This may be always?
            df_bayes.columns += ' interval'
        if not add_data:  # Done
            return df_bayes, trace
        return hdi2df_one_condition(df_bayes, posterior_index_name, df_data), trace
    else:
        est = xr.DataArray([max_a_p],
                           coords={'hdi': ["center"], posterior_index_name: hdi[posterior_index_name]},
                           dims=['hdi', posterior_index_name])

        df_bayes = xr.concat([hdi, est], 'hdi').rename('interval').to_dataframe()
        df_bayes = df_bayes.pivot_table(index=posterior_index_name, columns=['hdi', ]).reset_index()
        # Reset 2-level column from pivot_table:
        df_bayes.columns = [" ".join(np.flip(pair)) for pair in df_bayes.columns]
        df_bayes.rename({f' {posterior_index_name}': posterior_index_name}, axis=1, inplace=True)
        if not add_data:  # Done
            return df_bayes, trace
        return hdi2df_many_conditions(df_bayes, posterior_index_name, df_data), trace


def make_fold_change(df, y='log_firing_rate', index_cols=('Brain region', 'Stim phase'),
                     treatment_name='stim', treatments=(0, 1), do_take_mean=False, fold_change_method='divide'):
    for treatment in treatments:
        assert treatment in df[treatment_name].unique(), f'{treatment} not in {df[treatment_name].unique()}'
    if y not in df.columns:
        raise ValueError(f'{y} is not a column in this dataset: {df.columns}')
    index_cols = list(index_cols)
    # Take mean of trials:
    if do_take_mean:
        df = df.groupby(index_cols).mean().reset_index()

    # Make multiindex
    mdf = df.set_index(index_cols).copy()
    if (mdf.xs(treatments[1], level=treatment_name).size !=
        mdf.xs(treatments[0], level=treatment_name).size):
        raise IndexError(f'Uneven number of entries in conditions! Try setting do_take_mean=True'
                         f'{mdf.xs(treatments[0], level=treatment_name).size, mdf.xs(treatments[1], level=treatment_name).size}')

    # Subtract/divide
    try:
        if fold_change_method == 'subtract':
            data = (
                mdf.xs(treatments[1], level=treatment_name)[y] -
                mdf.xs(treatments[0], level=treatment_name)[y]
            ).reset_index()
        else:
            data = (mdf.xs(treatments[1], level=treatment_name)[y] /
                    mdf.xs(treatments[0], level=treatment_name)[y]
                    ).reset_index()
    except Exception as e:
        print(f'Try recasting {treatment_name} as integer and try again. Alternatively, use bayes_window.workflow.'
              f' We do that automatically there ')
        raise e
    y1 = f'{y} diff'
    data.rename({y: y1}, axis=1, inplace=True)
    if np.isnan(data[y1]).all():
        print(f'For {treatments}, data has all-nan {y1}: {data.head()}')
        print(f'Condition 1: {mdf.xs(treatments[1], level=treatment_name)[y].head()}')
        print(f'Condition 2: {mdf.xs(treatments[0], level=treatment_name)[y].head()}')
        raise ValueError(f'For {treatments}, data has all-nan {y1}. Ensure there a similar treatment to {y} does not'
                         f'shadow it!')

    return data, y1



from bayes_window.generative_models import generate_fake_spikes
from bayes_window.model_comparison import *
from pytest import mark


def test_run_methods():
    res = run_conditions(
        true_slopes=np.hstack([np.zeros(2), np.linspace(8.03, 18, 3)]),
        n_trials=[7],
        parallel=True
    )
    plot_roc(res)[0].display()
    plot_roc(res)[1].display()


@mark.parametrize('parallel', [False, True])
def test_compare_models(parallel):
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=10,
                                                                    n_neurons=10,
                                                                    n_mice=4,
                                                                    dur=2, )
    compare_models(df=df,
                   models={'no_neuron': models.model_hierarchical,
                           'no_neuron_or_teratment': models.model_hierarchical,
                           'no-treatment': models.model_hierarchical,
                           'treatment': models.model_hierarchical,
                           'student': models.model_hierarchical,
                           'lognogmal': models.model_hierarchical,

                           },
                   extra_model_args=[{'treatment': 'stim', 'condition': None},
                                     {'treatment': None, 'condition': None},
                                     {'treatment': None, 'condition': 'neuron'},
                                     {'treatment': 'stim', 'condition': 'neuron'},
                                     {'treatment': 'stim', 'condition': 'neuron', 'dist_y': 'student'},
                                     {'treatment': 'stim', 'condition': 'neuron', 'dist_y': 'lognormal'}, ],
                   y='isi',
                   group='mouse',
                   parallel=parallel,
                   plotose=True
                   )


@mark.parametrize('parallel', [False, True])
def test_compare_models2(parallel):
    df, df_monster, index_cols, _ = generate_fake_lfp(mouse_response_slope=13,
                                                      n_trials=40)
    compare_models(df=df,
                   models={
                       'no_teratment': models.model_hierarchical,
                       'no_group': models.model_hierarchical,
                       'full_normal': models.model_hierarchical,
                       'full_student': models.model_hierarchical,
                       'full_lognogmal': models.model_hierarchical,

                   },
                   extra_model_args=[
                       {'treatment': None},
                       {'group': None},
                       {'treatment': 'stim'},
                       {'treatment': 'stim', 'dist_y': 'student'},
                       {'treatment': 'stim', 'dist_y': 'lognormal'},
                   ],
                   y='isi',
                   condition=None,
                   parallel=False
                   )



from itertools import product

import altair as alt
import arviz as az
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from jax import random
from joblib import Parallel, delayed
from numpyro.infer import Predictive
from sklearn.metrics import confusion_matrix
from sklearn.metrics import roc_curve, auc
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import LabelEncoder
from tqdm import tqdm

from bayes_window import workflow, models
from bayes_window.fitting import fit_numpyro
from bayes_window.generative_models import generate_fake_lfp

trans = LabelEncoder().fit_transform


def make_confusion_matrix(res, groups):
    df = []
    for _, this_res in res.groupby(list(groups)):
        this_res['score'] = this_res['score'].replace({'': None}).astype(float)
        this_res['true_slope'] = this_res['true_slope'] > 0
        this_res['score'] = this_res['score'] > 0

        cm = confusion_matrix(this_res['true_slope'],
                              this_res['score'],
                              labels=this_res['true_slope'].unique())
        cm = [y for i in cm for y in i]
        roll = list(product(np.unique(this_res['true_slope']), repeat=2))
        this_res = this_res.drop(['true_slope', 'score'], axis=1)
        for i in range(len(roll)):
            rez = {'actual': roll[i][0],
                   'predicted': roll[i][1],
                   'Occurences': cm[i],
                   }
            rez.update(this_res.iloc[0].to_dict())
            df.append(rez)
        # Remove raw scores to reduce confusion
    return pd.DataFrame.from_records(df)


def plot_confusion(df):
    # plot
    base = alt.Chart(df)
    heat = base.mark_rect().encode(
        x="predicted",
        y="actual",
        color='Occurences:O'
    ).properties(width=180, height=180)
    # Configure text
    # Doesnt work without mean; mean is meaningless without groupby
    text = base.mark_text(baseline='middle').encode(
        text=alt.Text('mean(Occurences)', format=",.1f", ),
        x="predicted",
        y="actual",
        # color=alt.condition(
        #    alt.datum.Occurences > df['Occurences'].mean(),
        #    alt.value('black'),
        #    alt.value('white')
        # )
    )
    return heat


def make_roc_auc(res, binary=True, groups=('method', 'y', 'randomness', 'n_trials')):
    """ Vary as function of true_slope """
    df = []
    for head, this_res in res.groupby(list(groups)):
        this_res['score'] = this_res['score'].replace({'': None}).astype(float)
        if binary:
            this_res['score'] = this_res['score'] > 0
        else:
            this_res = this_res.dropna(subset=['score'])  # drop nans to prevent errors
        fprs = []
        tprs = []
        slopes = []
        # Loop over true_slopes

        for ts in this_res[this_res['true_slope'] > 0]['true_slope'].unique():
            # Select all no-slopes and this slope:
            x = this_res[(this_res['true_slope'] == 0) |
                         (this_res['true_slope'] == ts)]
            # Binarize:
            x['true_slope'] = x['true_slope'] > 0
            fpr, tpr, _ = roc_curve(x['true_slope'], x['score'])
            # print(f"Yes for {head} {round(ts, 2)}: {fpr, tpr}")
            # print(f"{x['true_slope'].values},\n {x['score'].values}")
            fprs.extend(fpr)
            tprs.extend(tpr)
            slopes.extend(np.repeat(ts, len(fpr)))
        rocs = {'False positive rate': fprs,
                'True positive rate': tprs,
                'AUC': round(auc(fpr, tpr), 5),
                'true_slope': slopes
                }

        rocs = pd.DataFrame(rocs)

        # Remove raw scores to reduce confusion
        this_res = this_res.drop(['true_slope', 'score'], axis=1)
        for col in this_res.columns:
            rocs[col] = this_res[col].iloc[0]
        df.append(rocs)
    return pd.concat(df)


def make_roc_auc_old(res, binary=True, groups=('method', 'y', 'randomness', 'n_trials')):
    # Make ROC and AUC
    df = []
    for _, this_res in res.groupby(list(groups)):
        this_res['score'] = this_res['score'].replace({'': None}).astype(float)
        this_res['true_slope'] = this_res['true_slope'] > 0
        if binary:
            this_res['score'] = this_res['score'] > 0
        else:
            this_res = this_res.dropna(subset=['score'])  # drop nans to prevent errors
        fpr, tpr, _ = roc_curve(this_res['true_slope'], this_res['score'])

        # Remove raw scores to reduce confusion
        this_res = this_res.drop(['true_slope', 'score'], axis=1)

        # Only keep the number of rows that will be useful for keeping ROC
        this_res = this_res.reset_index(drop=True).iloc[:len(fpr)]

        this_res['False positive rate'] = fpr
        this_res['True positive rate'] = tpr
        this_res['AUC'] = round(auc(fpr, tpr), 5)

        df.append(this_res)
    return pd.concat(df)


def plot_roc(df):
    roc = (alt.Chart(df).mark_line(size=2.6, opacity=.7).encode(
        x='False positive rate',
        y='mean(True positive rate)',
        color='method'
    ) + alt.Chart(df).mark_point(size=22.6, opacity=.7).encode(
        x='False positive rate',
        y='mean(True positive rate)',
        color='method'
    )).interactive().properties(width=150)
    bars = alt.Chart(df).mark_bar().encode(
        x='method',
        y='AUC',
        color='method'
    ).interactive()
    return bars, roc


def run_method(df, method='bw_student', y='Log power'):
    bw = workflow.BayesWindow(df, y=y, treatment='stim', group='mouse')
    if method[:2] == 'bw':
        bw.fit_slopes(model=models.model_hier_stim_one_codition,
                      dist_y=method[3:],
                      add_data=False)
        return bw.data_and_posterior['lower interval'].iloc[0]
    elif method[:5] == 'anova':
        return bw.fit_anova()  # Returns p-value

    elif method == 'mlm':
        posterior = bw.fit_lme(add_data=False).posterior
        try:
            return posterior['lower interval'].iloc[0]
        except AttributeError:
            return posterior['lower interval']


def run_methods(methods, ys, true_slope, n_trials, randomness, parallel=False):
    df, df_monster, index_cols, _ = generate_fake_lfp(mouse_response_slope=true_slope,
                                                      n_mice=6,
                                                      n_trials=n_trials,
                                                      trial_baseline_randomness=randomness
                                                      )
    if parallel:
        res = Parallel(n_jobs=len(methods) * len(ys), verbose=0)(delayed(run_method)(df=df, method=method, y=y)
                                                                 for y, method in product(ys, methods))
    else:
        res = [run_method(df=df, method=method, y=y) for y, method in tqdm(product(ys, methods))]
    # Save result in dict
    return pd.DataFrame.from_records([{'method': method, 'y': y, 'score': mres, 'true_slope': true_slope,
                                       'n_trials': n_trials, 'randomness': randomness}
                                      for (method, y), mres in zip(product(methods, ys), res)])


def run_conditions(true_slopes=np.hstack([np.zeros(180), np.linspace(.03, 18, 140)]),
                   n_trials=range(8, 30, 7),
                   trial_baseline_randomness=(.2, .4, 1.8),
                   parallel=False,
                   methods=('bw_lognormal', 'bw_normal', 'mlm', 'anova',),  # 'bw_student'
                   ys=('Log power', 'Power',)):
    conditions = list(product(true_slopes, n_trials, trial_baseline_randomness))
    if parallel:
        res = Parallel(n_jobs=12)(delayed(run_methods)(methods, ys, true_slope, n_trials, randomness, parallel=False)
                                  for true_slope, n_trials, randomness in tqdm(conditions))
    else:
        res = [run_methods(methods, ys, true_slope, n_trials, randomness, parallel=False)
               for true_slope, n_trials, randomness in tqdm(conditions)]

    return pd.concat(res)


def split_train_predict(df, model, y, **kwargs):
    """

        y=df['con_coherence_magnitude_near'].values,
        treatment=trans(df['event'].values),
        condition=trans(df['condition_code'].values),
        group=trans(df['subject'].values),
    :return:
    """
    # condition = condition if type(condition) == list else [condition]
    # if condition[0]:
    #     assert condition[0] in df.columns
    level_names = ['treatment', 'condition', 'group']
    level_names = [level_name for level_name in level_names if (level_name in kwargs.keys())]
    level_names = [level_name for level_name in level_names if kwargs[level_name] is not None]
    for level_name in level_names:
        print(level_name)
    # Corresponding df labels:
    from bayes_window import utils
    # df_cols = utils.parse_levels(*level_names)

    df_cols = [utils.level_to_data_column(level_name, kwargs) for level_name in level_names]

    # split into training and test
    if len(df_cols) > 0:
        df_train, df_test = train_test_split(df, train_size=.5, test_size=.5, stratify=df[df_cols])
    else:
        df_train, df_test = train_test_split(df, train_size=.5, test_size=.5)
    model_args = {'y': df_test[y].values}
    model_args.update({level: trans(df_test[kwargs[level]]) for level in level_names})

    mcmc = fit_numpyro(model=model, **model_args, convert_to_arviz=False)
    ppc = Predictive(model, parallel=False, num_samples=1000)
    ppc = ppc(random.PRNGKey(17), **model_args)

    predictive = az.from_numpyro(mcmc,
                                 posterior_predictive=ppc,
                                 # coords={"mouse": df_test['mouse_code']},
                                 # dims={"y": ["mouse"]},
                                 )

    return predictive


def r2(trace):
    """
    R squared
    :param trace:
    """
    y_true = trace.observed_data["y"].values
    y_pred = trace.posterior_predictive.stack(sample=("chain", "draw"))["y"].values.T
    try:
        print(az.r2_score(y_true, y_pred))
    except (TypeError, ValueError) as e:
        print(e)


def compare_models(df, models: dict,
                   extra_model_args: list = None,
                   parallel=False, plotose=False, **kwargs):
    """
    compare_models(models={'Hier':bayes.Numpyro.model_hier,
                           'Hier+covariance':bayes.Numpyro.model_hier_covar,
                           'Twostep Exponential':bayes.TwoStep.model_twostep,
                           'Twostep Gamma':bayes.TwoStep.model_twostep,
                          },
                   data=[df,df,df_monster,df_monster],
                   extra_args=[{}, {}, {'prior':'Exponential'}, {'prior':'Gamma'}])
    """

    # numpyro.set_host_device_count(10)
    extra_model_args = extra_model_args or np.tile({}, len(models))
    if parallel:
        traces = Parallel(n_jobs=len(models))(delayed(split_train_predict)
                                              (df, model, **kwargs, **extra_model_arg)
                                              for model, extra_model_arg in
                                              zip(models.values(), extra_model_args))
    else:
        traces = [split_train_predict(df, model, **kwargs, **extra_model_arg)
                  for model, extra_model_arg in zip(models.values(), extra_model_args)]

    # save tp dict
    traces_dict = {}  # initialize results
    for key, trace in zip(models.keys(), traces):
        traces_dict[key] = trace

    if plotose:
        for trace_name, trace in traces_dict.items():

            # Plot PPC
            az.plot_ppc(trace,
                        # flatten=[treatment],
                        # flatten_pp=data_cols[2],
                        mean=False,
                        # num_pp_samples=1000,
                        # kind='cumulative'
                        )
            plt.title(trace_name)
            plt.show()
            r2(trace)
            # Weird that r2=1
            # Waic
            try:
                print('======= WAIC (higher is better): =========')
                print(az.waic(trace, pointwise=True))
                print(az.waic(trace, var_name='y'))
            except TypeError:
                pass

        try:
            for trace_name in traces_dict.keys():
                trace = traces_dict[trace_name]
                # Print diagnostics and effect size
                print(f"n(Divergences) = {trace.sample_stats.diverging.sum(['chain', 'draw']).values}")
                try:
                    slope = trace.posterior['v_mu'].sel({'v_mu_dim_0': 1}).mean(['chain']).values
                except Exception:
                    slope = trace.posterior['b'].mean(['chain']).values
                print(f'Effect size={(slope.mean() / slope.std()).round(2)}  == {trace_name}')
        except Exception:
            pass

    model_compare = az.compare(traces_dict)  # , var_name='y')
    az.plot_compare(model_compare, textsize=12, show=True)
    return model_compare



import bulwark.checks as ck
from altair.vegalite.v4.api import FacetChart, Chart, LayerChart
from bayes_window import models, fake_spikes_explore
from bayes_window.fitting import fit_numpyro
from bayes_window.generative_models import generate_fake_spikes
from bayes_window.utils import add_data_to_posterior
from bayes_window.visualization import plot_data_slope_trials
from sklearn.preprocessing import LabelEncoder

trans = LabelEncoder().fit_transform


def test_fake_spikes_explore():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    charts = fake_spikes_explore(df, df_monster, index_cols)
    for chart in charts:
        assert ((type(chart) == FacetChart) |
                (type(chart) == Chart) |
                (type(chart) == LayerChart)), print(f'{type(chart)}')
        chart.display()


def test_plot_data_and_posterior():
    # Make some data
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )

    for y in (set(df.columns) - set(index_cols)):
        # Estimate model
        trace = fit_numpyro(y=df[y].values,
                            treatment=(df['stim']).astype(int).values,
                            condition=trans(df['neuron']),
                            group=trans(df['mouse']),
                            progress_bar=True,
                            model=models.model_hierarchical,
                            n_draws=100, num_chains=1, )

        # Add data back
        df_both, trace.posterior = add_data_to_posterior(df,
                                                         posterior=trace.posterior,
                                                         y=y,
                                                         fold_change_index_cols=['neuron', 'stim', 'mouse', ],
                                                         treatment_name='stim',
                                                         treatments=(0, 1),
                                                         b_name='b_stim_per_condition',  # for posterior
                                                         posterior_index_name='neuron'  # for posterior
                                                         )

        # Plot data and posterior
        # chart = plot_data_and_posterior(df=df_both, y=f'{y} diff', x='neuron', color='mouse', title=y)
        # assert ((type(chart) == FacetChart) |
        #         (type(chart) == Chart) |
        #         (type(chart) == LayerChart)), print(f'{type(chart)}')
        trace.posterior.to_dataframe().pipe(ck.has_no_nans)

        # chart.display()


def test_plot_data_slope_trials():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )
    chart = plot_data_slope_trials(df=df, x='stim:O', y='log_firing_rate', color='neuron:N', detail='i_trial')
    chart.display()



import warnings

import altair as alt
import numpy as np
import pandas as pd
import xarray as xr
from elephant.spike_train_generation import inhomogeneous_poisson_process as poisson
from neo import analogsignal
from quantities import s
from sklearn.preprocessing import LabelEncoder

from bayes_window import utils

trans = LabelEncoder().fit_transform
warnings.filterwarnings("ignore")


# fake data generation
def generate_fake_lfp(n_trials=10,
                      n_mice=10,
                      dur=7,
                      mouse_response_slope=1.,
                      **kwargs):
    # neuron becomes mouse to provide different baselines
    df, df_monster, index_cols, _ = generate_fake_spikes(n_trials=n_trials,
                                                         n_neurons=n_mice,
                                                         n_mice=2,
                                                         dur=dur,
                                                         mouse_response_slope=mouse_response_slope,
                                                         **kwargs)

    df = df[df.mouse_code == 1]
    df = df[df.neuron.astype(int) < n_mice - 2]
    df = df.drop('mouse', axis=1)
    # df_monster = df_monster.drop('mouse', axis=1)
    df = df.rename({'firing_rate': 'Power', 'log_firing_rate': 'Log power', 'neuron': 'mouse'}, axis=1)
    # df_monster = df_monster.rename({'firing_rate': 'Power', 'log_firing_rate': 'Log power', 'neuron': 'mouse'}, axis=1)
    return df, None, index_cols, None


def generate_spikes_stim_types(mouse_response_slope=3, **kwargs):
    df1 = generate_fake_spikes(mouse_response_slope=mouse_response_slope, **kwargs)[0]
    df1.insert(0, 'stim_strength', np.ones(df1.shape[0]))
    df2 = generate_fake_spikes(mouse_response_slope=2 * mouse_response_slope, **kwargs)[0]
    df2.insert(0, 'stim_strength', 2 * np.ones(df1.shape[0]))
    return pd.concat([df1, df2])


def generate_fake_spikes(n_trials=6,
                         n_neurons=8,
                         n_mice=10,
                         dur=7,
                         slowest_fr=7,
                         time_randomness=10,
                         overall_stim_response_strength=0,
                         mouse_response_slope=2.,  # .3 is hard, 3 is easy
                         verbose=False,
                         do_ave_trial=False,
                         trial_baseline_randomness=.2,
                         do_bad_mice=False,
                         ):
    """
    # mouse id affects slope of fr
    # Stim affects Firing rates for slow more than fast neurons

    """
    trans = LabelEncoder().fit_transform

    def firing_rate_over_time(rate, rate0=None, dur=10, time_randomness=None):
        time_rate = np.tile(rate, dur)
        time_rate[time_rate < 0] = 0.01  # check that no zeros
        #  add irrelevant blip of fr at end of spiketrain
        # Only 1/4 is relevant. Out of 7 sec, that's 7*0.25=1.75 sec
        if not (rate0 is None):
            time_rate[:int(dur / 4 * 2)] = rate0
            time_rate[-int(dur / 4):] = rate0
        return time_rate + time_randomness * np.random.random(dur)

    # Response strength varies by mouse:
    stim_response_mouse = overall_stim_response_strength + np.arange(n_mice) * mouse_response_slope
    if do_bad_mice:  # Half the mice have no response to stim at all
        stim_response_mouse[:int(n_mice / 2)] = 0
    if verbose:
        print(f'Stim response per mouse (in Hz) is {stim_response_mouse}')

    fast_rates = np.logspace(np.log10(slowest_fr + .1 + stim_response_mouse[range(n_mice)]),
                             np.log10(38), n_neurons)
    slow_rates = np.logspace(np.tile(np.log10(slowest_fr), n_mice),
                             np.log10(38), n_neurons)
    if verbose:
        print(f'Stim response per mouse (in Hz) is {fast_rates - slow_rates}')
        print(f'Fast rates = {fast_rates},\n slow rates= {slow_rates}')

    # make an Xarray of firing rates:
    slow = xr.DataArray(np.vstack(slow_rates).T,
                        dims=['Mouse', 'Neuron'],
                        coords=dict(Mouse=[f'm{i}bayes' for i in range(n_mice)],
                                    Neuron=range(n_neurons)), name='Firing rate')
    fast = xr.DataArray(np.vstack(fast_rates).T,
                        dims=['Mouse', 'Neuron'],
                        coords=dict(Mouse=[f'm{i}bayes' for i in range(n_mice)],
                                    Neuron=range(n_neurons)), name='Firing rate')
    slow['Stim'] = False
    fast['Stim'] = True
    firing_rates = xr.concat([slow, fast], dim='Stim')

    ddf = []
    for (imouse, i_neuron), firing_rate in firing_rates.to_dataframe().groupby(['Mouse', 'Neuron']):

        firing_rate.reset_index(inplace=True)
        fast_rate = firing_rate[firing_rate['Stim']]['Firing rate'].values
        slow_rate = firing_rate[~firing_rate['Stim']]['Firing rate'].values
        for i_trial in range(n_trials):
            # generate  independent Poisson spike trains that varies with time
            spiketrains_stim = np.array(poisson(
                analogsignal.AnalogSignal(
                    firing_rate_over_time(fast_rate,
                                          slow_rate + abs(np.random.randn() * trial_baseline_randomness),
                                          dur=dur,
                                          time_randomness=time_randomness),
                    1 / s, sampling_rate=1 / s)))
            spiketrains_nostim = np.array(poisson(
                analogsignal.AnalogSignal(
                    firing_rate_over_time(slow_rate, dur=dur, time_randomness=time_randomness),
                    1 / s, sampling_rate=1 / s)))
            for i_spike, spike in enumerate(spiketrains_stim):
                ddf.append(dict(spike_time=spike,
                                neuron=str(i_neuron),
                                stim=True,
                                mouse=firing_rate['Mouse'].values[0],
                                i_trial=i_trial,
                                i_spike=i_spike))

            for i_spike, spike in enumerate(spiketrains_nostim):
                ddf.append(dict(spike_time=spike,
                                neuron=str(i_neuron),
                                stim=False,
                                mouse=firing_rate['Mouse'].values[0],
                                i_trial=i_trial,
                                i_spike=i_spike))

    df = pd.DataFrame.from_records(ddf).sort_values(
        ['neuron', 'stim', 'mouse', 'i_trial', 'spike_time']).reset_index(drop=True)

    # Make index for pymc

    df['neuron_x_mouse'] = df['neuron'] + df['mouse']
    df['mouse_code'] = trans(df['mouse'])
    df['neuron_code'] = trans(df['neuron_x_mouse'])
    df['mouse_code'] = trans(df['mouse'])
    df['stim'] = trans(df['stim'])
    index_cols = set(df.columns) - {'spike_time', 'firing_rate', 'Firing rate', 'log_firing_rate', 'num_spikes', 'isi',
                                    'log_isi', '1/isi', 'log_1/isi'}

    # make full ISI
    df['isi'] = np.diff(df['spike_time'], prepend=0)
    df = df[df['isi'] > 0]  # drop the first spike of every trial, since impossible to estimate its isi
    df = df.drop(0)  # drop the first spike, since impossible to estimate its isi
    df['1/isi'] = df['isi'].apply(lambda x: 1 / x)
    df['log_1/isi'] = np.log10(df['1/isi'])
    df['log_isi'] = np.log10(df['isi'])

    if do_ave_trial:
        df = df.groupby(['neuron', 'neuron_x_mouse', 'mouse', 'stim']).mean().reset_index()

    # Trialwise measures of firing rate:
    dddf = []
    for i, ddf in df.groupby(list(index_cols - {'i_spike'})):
        ddf.insert(0, 'firing_rate', ddf['1/isi'].mean())
        ddf.insert(0, 'Firing rate', ddf['firing_rate'])
        ddf.insert(0, 'log_firing_rate', np.log10(ddf['firing_rate']))
        dddf.append(ddf)

    df = pd.concat(dddf)

    df_monster = df.copy()

    # make reduced dataframe
    # Mean over rounds, drop spike time and other stuff that doesnt make sense for single-trial data
    df = df.set_index(['i_trial']).groupby(list(index_cols - {'i_spike'})).mean().drop(
        ['log_isi', '1/isi', 'log_1/isi', 'spike_time'], axis=1).reset_index()

    if verbose:
        print(f'Difference in firing rates is {fast_rates[0] - slow_rates[0]}')
    return df, df_monster, index_cols, firing_rates


def fake_spikes_explore(df, df_monster, index_cols):
    # mean firing rate per trial per mouse
    width = 50
    fig_trials = alt.Chart(df).mark_line(fill=None, ).encode(
        x=alt.X('stim'),
        y=alt.Y('mean(log_firing_rate)', scale=alt.Scale(zero=False)),
        color='neuron:N',
        detail='i_trial:Q',
        opacity=alt.value(1),
        size=alt.value(.9),
        facet='mouse'
    ).properties(
        title='All trials and neurons',
        # columns=5,
        width=width,
        height=300
    )

    # mean firing rate per trial per mouse (select first and last mouse)
    alt.data_transformers.disable_max_rows()
    fig_select = alt.Chart(df[(df['neuron'] == '0') |
                              (df['neuron'] == str(df['neuron'].astype(int).max().astype(int) - 1))]).mark_line(
        fill=None, ).encode(
        x=alt.X('stim'),
        y=alt.Y('mean(log_firing_rate)', scale=alt.Scale(zero=False)),
        color='neuron:N',
        detail='i_trial:Q',
        opacity=alt.value(1),
        size=alt.value(2),
        facet='mouse'
    ).properties(
        title='Slow neurons are more responsive',
        # columns=5,
        width=width,
        height=300
    )

    # mean firing rate per mouse
    fig_neurons = alt.Chart(df).mark_line(fill=None, ).encode(
        x=alt.X('stim'),
        y=alt.Y('mean(log_firing_rate)', scale=alt.Scale(zero=False)),
        color='neuron:N',
        opacity=alt.value(1),
        size=alt.value(2),
        facet='mouse'
    ).properties(
        title='All neurons',
        # columns=5,
        width=width,
        height=300
    )
    # mean firing rate per mouse
    fig_mice = alt.Chart(df[df['neuron'] == '0']).mark_line(fill=None, ).encode(
        x=alt.X('stim'),
        y=alt.Y('mean(log_firing_rate)', scale=alt.Scale(zero=False)),
        opacity=alt.value(1),
        size=alt.value(3),
        facet='mouse:N'
    ).properties(
        title='Mice sorted by response',
        # columns=5,
        width=width,
        height=300
    )

    # Monster-level ISI
    df_isi = df_monster[
        (
            (df_monster['neuron'] == '0') |
            (df_monster['neuron'] == str(df_monster['neuron'].astype(int).max()))
        ) &
        # (df_monster['mouse']=='m0bayes') |
        (df_monster['mouse'] == f'm{df_monster["mouse_code"].astype(int).max()}bayes')
        ]
    fig_isi = alt.Chart(df_isi).mark_tick(opacity=.2).encode(
        x=alt.X('stim'),
        y=alt.Y('mean(log_1/isi)', scale=alt.Scale(zero=False), ),
        color='neuron:N',
        detail='i_spike:Q',  # Crucial: Q!
    ).properties(
        title=['Multiple trials per mouse', 'many spikes'],
        height=500
    )
    fig_overlay = alt.Chart(df_isi).mark_line(fill=None, ).encode(
        x=alt.X('stim'),
        y=alt.Y('log_firing_rate', scale=alt.Scale(zero=False)),
        color='neuron:N',
        detail='i_trial:Q',
        size=alt.value(2),
    )

    data_fold_change, y = utils.make_fold_change(df, y='log_firing_rate',
                                                 index_cols=list(set(index_cols) - {'i_spike'}),
                                                 treatment_name='stim',
                                                 treatments=(0, 1))
    box = alt.Chart(data=data_fold_change).mark_boxplot().encode(y=y).encode(
        x=alt.X('neuron:N', ),
        y=alt.Y(y, scale=alt.Scale(zero=True)),
    ).properties(width=width, height=240).facet(
        # row='mouse:N',
        column=alt.Column('mouse'))

    bar = (alt.Chart(data=data_fold_change).mark_bar().encode(y=alt.Y(y, aggregate='mean')) +
           alt.Chart(data=data_fold_change).mark_errorbar().encode(y=alt.Y(y, aggregate='stderr'))).encode(
        x=alt.X('neuron:N', ),
        y=alt.Y(y),
    ).properties(width=width * 2, height=240).facet(
        # row='Inversion:N',
        column=alt.Column('mouse'))

    bar_combined = (alt.Chart(data=data_fold_change).mark_bar().encode(y=alt.Y(y, aggregate='mean')) +
                    alt.Chart(data=data_fold_change).mark_errorbar().encode(y=alt.Y(y, aggregate='stderr'))).encode(
        x=alt.X('neuron:N', ),
        y=alt.Y(y),
    ).properties(width=width, height=240)  # .facet(
    # row='Inversion:N',
    # column=alt.Column('mouse'))#.resolve_scale(y='independent')

    # Monster-level ISI
    df_raster = df_monster[
        # (
        #    (df_monster['neuron']=='0')       |
        #    (df_monster['neuron']==str(n_neurons-1))
        # )
        # &
        # (df_monster['mouse']==f'm{df_monster["mouse_code"].astype(int).max()}bayes') |
        (df_monster['mouse'] == f'm{df_monster["mouse_code"].astype(int).max()}bayes')
    ]
    fig_raster = alt.Chart(df_raster).mark_tick(thickness=.8).encode(
        y=alt.Y('neuron'),
        x=alt.X('spike_time', scale=alt.Scale(zero=False), ),
        # color='neuron:N',
        detail='i_spike:Q',
    ).properties(
        # title=['Multiple trials per mouse','many spikes'],
        width=800,
        height=140
    ).facet(row='stim')

    return fig_mice, fig_select, fig_neurons, fig_trials, fig_isi + fig_overlay, bar, box, fig_raster, bar_combined



import bulwark.checks as ck
import xarray as xr
from bayes_window import models
from bayes_window.fitting import fit_numpyro
from bayes_window.generative_models import generate_fake_spikes
from joblib import delayed, Parallel
from sklearn.preprocessing import LabelEncoder

trans = LabelEncoder().fit_transform


def test_fit_numpyro_serial():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes(n_trials=2,
                                                                    n_neurons=3,
                                                                    n_mice=4,
                                                                    dur=2, )

    for y in (set(df.columns) - set(index_cols)):
        trace = fit_numpyro(y=df[y].values,
                            treatment=trans(df['stim']),
                            condition=trans(df['neuron']),
                            group=trans(df['mouse']),
                            progress_bar=True,
                            model=models.model_hierarchical,
                            n_draws=100, num_chains=1, )
        # chart = visualization.plot_posterior_altair(trace,
        #                                             df,
        #                                             b_name='b_stim_per_condition',
        #                                             x='stim',
        #                                             group_name='neuron'
        #                                             )
        #
        # assert ((type(chart) == FacetChart) |
        #         (type(chart) == Chart) |
        #         (type(chart) == LayerChart)), print(f'{type(chart)}')
        trace.posterior.to_dataframe().pipe(ck.has_no_nans)


def test_fit_numpyro_parallel():
    df, df_monster, index_cols, firing_rates = generate_fake_spikes()
    # df = df.set_index(list(index_cols))
    meas = (set(df.columns) - set(index_cols))

    traces = Parallel(n_jobs=-1, verbose=2,
                      # backend='multiprocessing'
                      )(
        delayed(fit_numpyro)(y=y,
                             treatment=(df['stim']).astype(int).values,
                             condition=trans(df['neuron']),
                             group=trans(df['mouse']),
                             progress_bar=False,
                             model=models.model_hierarchical,
                             n_draws=10
                             )
        for y in [df[y].values for y in meas])
    assert type(traces[0].posterior) == xr.Dataset
    [trace.posterior.to_dataframe().pipe(ck.has_no_nans) for trace in traces]



"""Top-level package ."""

__author__ = """Maxym Myroshnychenko"""
__email__ = 'mmyros@gmail.com'
__version__ = '0.1.0'

from .bayes_window import *
from .workflow import *
from .generative_models import *
from .fitting import *
from .model_comparison import *
from .utils import *
from .visualization import *



import arviz as az
import numpyro
from jax import random
from numpyro.infer import MCMC, NUTS

from . import models


def fit_numpyro(progress_bar=False, model=None, n_draws=1000, num_chains=1, convert_to_arviz=True, **kwargs):
    model = model or models.model_hierarchical
    numpyro.set_host_device_count(4)
    mcmc = MCMC(NUTS(model), num_warmup=1000, num_samples=n_draws, num_chains=num_chains, progress_bar=progress_bar)
    mcmc.run(random.PRNGKey(16), **kwargs)

    # arviz convert
    trace = az.from_numpyro(mcmc)
    # Print diagnostics
    if trace.sample_stats.diverging.sum(['chain', 'draw']).values > 0:
        print(f"n(Divergences) = {trace.sample_stats.diverging.sum(['chain', 'draw']).values}")
    if convert_to_arviz:
        return trace
    else:
        return mcmc



import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist


def model_single_normal_stim(y, treatment, condition):
    n_conditions = np.unique(condition).shape[0]
    a = numpyro.sample('a', dist.Normal(0, 1))
    b = numpyro.sample('b', dist.Normal(0, 1))
    sigma_b = numpyro.sample('sigma_b', dist.HalfNormal(1))
    b_stim_per_condition = numpyro.sample('b_stim', dist.Normal(jnp.tile(b, n_conditions), sigma_b))
    theta = a + b_stim_per_condition[condition] * treatment

    sigma_obs = numpyro.sample('sigma_obs', dist.HalfNormal(1))
    numpyro.sample('y', dist.Normal(theta, sigma_obs), obs=y)


def sample_y(dist_y, theta, sigma_obs, y):
    if dist_y == 'student':
        nu_y = numpyro.sample('nu_y', dist.Gamma(1, .1))
        numpyro.sample('y', dist.StudentT(nu_y, theta, sigma_obs), obs=y)
    elif dist_y == 'normal':
        numpyro.sample('y', dist.Normal(theta, sigma_obs), obs=y)
    elif dist_y == 'lognormal':
        numpyro.sample('y', dist.LogNormal(theta, sigma_obs), obs=y)
    else:
        raise ValueError


def model_single(y, condition, dist_y='normal'):
    n_conditions = np.unique(condition).shape[0]
    a_neuron = numpyro.sample('mu', dist.Normal(0, 1))
    sigma_neuron = numpyro.sample('sigma', dist.HalfNormal(1))
    a_neuron_per_condition = numpyro.sample('mu_per_condition',
                                            dist.Normal(jnp.tile(a_neuron, n_conditions), sigma_neuron))
    theta = a_neuron + a_neuron_per_condition[condition]
    sigma_obs = numpyro.sample('sigma_obs', dist.HalfNormal(1))
    sample_y(dist_y=dist_y, theta=theta, sigma_obs=sigma_obs, y=y)


def model_hierarchical(y, condition=None, group=None, treatment=None, dist_y='normal'):
    n_conditions = np.unique(condition).shape[0]
    n_subjects = np.unique(group).shape[0]
    a = numpyro.sample('a', dist.Normal(0, 1))

    # b_subject = numpyro.sample('b_subject', dist.Normal(jnp.tile(0, n_subjects), 1))
    # sigma_b_subject = numpyro.sample('sigma_b_subject', dist.HalfNormal(1))

    a_subject = numpyro.sample('a_subject', dist.Normal(jnp.tile(0, n_subjects), 1))
    sigma_a_subject = numpyro.sample('sigma_a_subject', dist.HalfNormal(1))
    if condition is None:
        b = numpyro.sample('b', dist.Normal(0, 1))
    sigma_b_condition = numpyro.sample('sigma_b_condition', dist.HalfNormal(1))
    b_stim_per_condition = numpyro.sample('b_stim_per_condition', dist.Normal(jnp.tile(0, n_conditions), .5))
    if group is not None:
        theta = a + a_subject[group] * sigma_a_subject
    else:
        theta = a
    if condition is not None:
        slope = (  # b + b_subject[subject] * sigma_b_subject
            + b_stim_per_condition[condition] * sigma_b_condition
        )
    else:
        slope = b
    if treatment is not None:
        slope = slope * treatment
    theta += slope
    sigma_obs = numpyro.sample('sigma_obs', dist.HalfNormal(1))
    sample_y(dist_y=dist_y, theta=theta, sigma_obs=sigma_obs, y=y)


def model_hier_stim_one_codition(y, treatment=None, group=None, dist_y='normal', **kwargs):
    n_subjects = np.unique(group).shape[0]
    a = numpyro.sample('a', dist.Normal(0, 1))

    # b_subject = numpyro.sample('b_subject', dist.Normal(jnp.tile(0, n_subjects), 1))
    # sigma_b_subject = numpyro.sample('sigma_b_subject', dist.HalfNormal(1))

    a_subject = numpyro.sample('a_subject', dist.Normal(jnp.tile(0, n_subjects), 1))
    sigma_a_subject = numpyro.sample('sigma_a_subject', dist.HalfNormal(1))

    b = numpyro.sample('b_stim_per_condition', dist.Normal(0, 1))

    theta = a + a_subject[group] * sigma_a_subject
    slope = b
    if treatment is not None:
        slope = slope * treatment
    theta += slope

    sigma_obs = numpyro.sample('sigma_obs', dist.HalfNormal(1))
    sample_y(dist_y=dist_y, theta=theta, sigma_obs=sigma_obs, y=y)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = []

setup_requirements = ['pytest-runner', 'joblib', 'numpy', 'numpyro', 'bulwark', 'elephant', 'xarray', 'altair',
                      # 'jax',
                      # 'jaxlib==0.1.51',
                      'pandas', 'arviz', 'sklearn', 'inflection','nbsphinx','nbval','statsmodels']

test_requirements = ['pytest>=3', ]

setup(
    author="Maxym Myroshnychenko",
    author_email='mmyros@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="Pretty and easy Bayesian estimation with data overlay",
    entry_points={
        'console_scripts': [
            'bayes_window=bayes_window.cli:main',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='bayes_window',
    name='bayes_window',
    packages=find_packages(include=['bayes_window', 'bayes_window.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/mmyros/bayes_window',
    version='0.1.0',
    zip_safe=False,
)



import altair as alt
import numpy as np
from sklearn.preprocessing import LabelEncoder

trans = LabelEncoder().fit_transform


def facet(base_chart,
          column=None,
          row=None,
          width=80,
          height=150,
          finalize=False
          ):
    if column is None and row is None:
        raise RuntimeError('Need either column, or row, or both!')
    assert base_chart.data is not None
    if column:
        assert column in base_chart.data.columns, f'{column} is not in {base_chart.data.columns}'
        # sanitize a little:
        base_chart.data[column] = base_chart.data[column].astype(str)
    if row:
        assert row in base_chart.data.columns, f'{row} is not in {base_chart.data.columns}'
        # sanitize a little:
        base_chart.data[row] = base_chart.data[row].astype(str)

    def concat_charts(subdata, groupby_name, row_name='', row_val='', how='hconcat'):

        charts = []
        for i_group, group_val in enumerate(subdata[groupby_name].drop_duplicates().sort_values()):
            title = f"{groupby_name} {group_val} {row_name} {row_val}"  # if i_group == 0 else ''
            charts.append(alt.layer(
                base_chart,
                title=title,
                data=base_chart.data
            ).transform_filter(
                alt.datum[groupby_name] == group_val
            ).resolve_scale(
                y='independent'
            ).properties(
                width=width, height=height))
        if how == 'hconcat':
            return alt.hconcat(*charts)
        else:
            return alt.vconcat(*charts)

    if row is None:
        chart = concat_charts(base_chart.data, groupby_name=column,
                              how='hconcat')
    elif column is None:
        chart = concat_charts(base_chart.data, groupby_name=row,
                              how='vconcat')
    else:
        chart = alt.vconcat(*[concat_charts(subdata, groupby_name=column,
                                            row_name=row, row_val=val, how='hconcat')
                              for val, subdata in base_chart.data.groupby(row)])
    if finalize:  # TODO is there a way to do this in theme instead?
        chart = chart.configure_view(
            stroke=None
        )
    return chart


def plot_data(df=None, x=None, y=None, color=None, add_box=True, base_chart=None, **kwargs):
    color = color or ':O'
    assert (df is not None) or (base_chart is not None)
    if (x == '') or (x[-2] != ':'):
        x = f'{x}:O'
    # Plot data:
    base = base_chart or alt.Chart(df)
    if color[-2] != ':':
        color = f'{color}:N'
    charts = []
    axis = alt.Axis()

    if (x != ':O') and (len(base.data[x[:-2]].unique()) > 1):
        charts.append(base.mark_line(clip=True, fill=None, opacity=.5, size=2.5).encode(
            x=x,
            color=f'{color}',
            y=alt.Y(f'mean({y})',
                    scale=alt.Scale(zero=False,
                                    domain=list(np.quantile(base.data[y], [.05, .95])))),
        ))
        axis = alt.Axis(labels=False, tickCount=0, title='')
    if add_box:
        # Shift x axis for box so that it doesnt overlap:
        # df['x_box'] = df[x[:-2]] + .01
        charts.append(base.mark_boxplot(clip=True, opacity=.3, size=12, color='black').encode(
            x=x,
            y=alt.Y(f'{y}:Q',
                    scale=alt.Scale(zero=False,
                                    domain=list(np.quantile(base.data[y], [.05, .95]))),
                    axis=axis)
        ))
    return alt.layer(*charts)


# from altair.vegalite.v4.api import Undefined
def plot_posterior(df=None, title='', x=':O', do_make_change=True, base_chart=None, **kwargs):
    assert (df is not None) or (base_chart is not None)
    data = base_chart.data if df is None else df
    if x[-2] != ':':
        x = f'{x}:O'  # Ordinal
    assert 'higher interval' in data.columns
    assert 'lower interval' in data.columns
    assert 'center interval' in data.columns
    base_chart = base_chart or alt.Chart(data=data)

    # line
    chart = base_chart.mark_line(clip=True, point=True, color='black', fill=None).encode(
        y=alt.Y('center interval:Q', impute=alt.ImputeParams(value='value')),
        x=x,
    )
    do_make_change = do_make_change is not False

    # Axis limits
    minmax = [float(data['lower interval'].min()), 0,
              float(data['higher interval'].max())]
    scale = alt.Scale(zero=do_make_change,  # Any string or True
                      domain=[min(minmax), max(minmax)])

    # Make the zero line
    if do_make_change:
        base_chart.data['zero'] = 0
        chart += base_chart.mark_rule(color='black', size=.1, opacity=.6).encode(y='zero')
        title = f'Δ {title}'

    # error_bars
    chart += base_chart.mark_rule().encode(
        x=x,
        y=alt.Y('lower interval:Q',
                title=title, scale=scale),
        y2='higher interval:Q',
    )

    return chart


def plot_data_slope_trials(x,
                           y,
                           detail,
                           color=None,
                           base_chart=None,
                           df=None,
                           **kwargs):
    assert (df is not None) or (base_chart is not None)
    color = color or ':O'
    if x[-2] != ':':
        x = f'{x}:O'  # Ordinal
    base_chart = base_chart or alt.Chart(df)

    # mean firing rate per trial per mouse
    fig_trials = base_chart.mark_line(fill=None).encode(
        x=alt.X(x),
        y=alt.Y(y, scale=alt.Scale(zero=False,
                                   # domain=[df[y].min(), df[y].max()])),
                                   domain=list(np.quantile(base_chart.data[y], [.05, .95])),
                                   clamp=True)),
        color=color,
        detail=detail,
        opacity=alt.value(.2),
        size=alt.value(.9),
        # **kwargs
    )
    return fig_trials
